.. _changelog:

=========
CHANGELOG
=========

{%    if ctx.changelog_mode == "update"
%}{#  # IMPORTANT: add insertion flag for next version update
#}{{
        insertion_flag ~ "\n"

}}{%  endif
%}
